import os

__license__ = "MIT"
__version__ = "2.1.0"
__author__ = "Zachary Zhang"
__email__ = "zlj19971222@outlook.com"
__git_url__ = "https://github.com/zlj-zz/fungit.git"
__HOME__ = os.environ["HOME"]
__FUNGITDIR__ = __HOME__ + "/.config/fungit"



import argparse
import signal

from ..logutil import setup_logging
from . import __TERLOG__
from .gitoptions import process
from .msg_helper import echo_help_msg, echo_description, echo_types, give_tip
from .shell_complete import add_completion
from .shared import exit_


def command_g(custom_commands: list = None):
    setup_logging(debug=False, log_file=__TERLOG__)

    try:
        signal.signal(signal.SIGINT, exit_)
    except Exception:
        pass

    args = argparse.ArgumentParser(
        prog="g",
        description="If you want to use some original git commands, please use -- to indicate.",
    )
    args.add_argument(
        "-c", "--complete", action="store_true", help="Add shell prompt script and exit"
    )
    args.add_argument(
        "-s",
        "--show-commands",
        action="store_true",
        help="List all available fame and wealth and exit",
    )
    args.add_argument(
        "-S",
        "--show-command",
        type=str,
        metavar="TYPE",
        dest="command_type",
        help="According to given type list available fame and wealth and exit",
    )
    args.add_argument(
        "-t",
        "--types",
        action="store_true",
        help="List all command type and exit",
    )
    args.add_argument(
        "command", nargs="?", default="|", type=str, help="Short git command"
    )
    args.add_argument("args", nargs="*", type=str, help="Command parameter list")
    stdargs = args.parse_args()

    if custom_commands is not None:
        stdargs = args.parse_args(custom_commands)
    # print(stdargs)

    if stdargs.complete:
        add_completion()
        raise SystemExit(0)

    if stdargs.show_commands:
        echo_help_msg()
        raise SystemExit(0)

    if stdargs.command_type:
        give_tip(stdargs.command_type)
        raise SystemExit(0)

    if stdargs.types:
        echo_types()
        raise SystemExit(0)

    if stdargs.command:
        if stdargs.command == "|":
            echo_description()
        else:
            command = stdargs.command
            process(command, stdargs.args)



from typing import List, Any

from fungit.style import Fx, Cursor
from ..shared import BoxType
from ..renderer import Renderer
from ..theme import Theme
from ..utils import create_profile
from . import Box
from .content_box import ContentBox


class NavBox(Box):
    # only change in father class, sub class only read
    current: int = BoxType.STATUS

    name: str
    genre: int
    x: int
    y: int
    w: int
    h: int
    raw: Any
    content: List
    selected: int
    box: str
    box_content: str

    @classmethod
    def set_current(cls, t: int):
        cls.current = t

    @classmethod
    def create_profile(cls):
        _line_color = Theme.BOX_SELECTED_COLOR if cls.genre & cls.current else ""
        _item_msg = (
            f"{cls.selected + 1} of {len(cls.content)}" if len(cls.content) > 1 else ""
        )
        cls.box = create_profile(
            cls.x, cls.y, cls.w, cls.h, cls.name, _item_msg, line_color=_line_color
        )

    @classmethod
    def fetch_data(cls):
        """Get raw data.

        Get the list of original data and cache it in `cls.raw`, each of which
        is an independent data object instance(Override by subclass).
        """
        pass

    @classmethod
    def generate(cls):
        """Content of production display.

        All the acquired data is produced as the class content that can be
        displayed, colored and cached in `cls.content`(Override by subclass).
        """
        pass

    @classmethod
    def update(cls):
        """Update display content.

        According to the currently selected item to update the displayed
        content block (current item BOLD), generate a content string, and
        assign it to `cla.box_content`.
        """
        start_x = cls.x + 1
        start_y = cls.y + 1
        line_w = cls.w - 2

        cls.box_content = ""
        for idx, line in enumerate(cls.content):
            if idx < cls.h - 2:
                _line = f"{Cursor.to(start_y, start_x)}{line if len(line) < line_w else line[:line_w]}"
                if cls.genre & cls.current:
                    _line = f"{Fx.b}{_line}{Fx.ub}"
                cls.box_content += _line
                start_y += 1

    @classmethod
    def render(cls):
        Renderer.now(cls.box, cls.box_content)

        if cls.current & cls.genre:
            ContentBox.notify(cls)

    @classmethod
    def notify(cls, update_data: bool = False, re_profile: bool = True):
        if update_data:
            cls.fetch_data()
            cls.generate()
        if re_profile:
            cls.create_profile()
        # every notify must update display string and render all.
        cls.update()
        cls.render()

    @classmethod
    def set_selected(cls, v):
        cls.selected = v
        cls.notify()



import re


class Fx:
    """Text effects
    * trans(string: str): Replace whitespace with escape move right to not overwrite background behind whitespace.
    * uncolor(string: str) : Removes all 24-bit color and returns string .
    """

    start = "\033["  # * Escape sequence start
    sep = ";"  # * Escape sequence separator
    end = "m"  # * Escape sequence end
    # * Reset foreground/background color and text effects
    reset = rs = "\033[0m"
    bold = b = "\033[1m"  # * Bold on
    unbold = ub = "\033[22m"  # * Bold off
    dark = d = "\033[2m"  # * Dark on
    undark = ud = "\033[22m"  # * Dark off
    italic = i = "\033[3m"  # * Italic on
    unitalic = ui = "\033[23m"  # * Italic off
    underline = u = "\033[4m"  # * Underline on
    ununderline = uu = "\033[24m"  # * Underline off
    blink = bl = "\033[5m"  # * Blink on
    unblink = ubl = "\033[25m"  # * Blink off
    strike = s = "\033[9m"  # * Strike / crossed-out on
    unstrike = us = "\033[29m"  # * Strike / crossed-out off

    # * Precompiled regex for finding a 24-bit color escape sequence in a string
    color_re = re.compile(r"\033\[\d+;\d?;?\d*;?\d*;?\d*m")

    @staticmethod
    def trans(string: str):
        return string.replace(" ", "\033[1C")

    @classmethod
    def uncolor(cls, string: str) -> str:
        return f'{cls.color_re.sub("", string)}'



from typing import List, Any

from fungit.commands import info
from fungit.gui.theme import Theme
from fungit.style import Cursor, Fx
from ..shared import BoxType
from . import Box

ADDITION_FLAG = "+"
DELETION_FLAG = "-"


class ContentBox(Box):
    name: str = ""
    genre = BoxType.CONTENT
    x: int = 0
    y: int = 0
    w: int = 0
    h: int = 0
    raw: Any = None
    content: List = []
    box: str = ""
    box_content: str = ""
    content_selected_line: int = 0

    box_selected_idx: int = -1  # record nav box selected item
    last_notifier: Box = None

    @classmethod
    def fetch_data(cls, box):
        cls.raw = fetch_content(box)

    @classmethod
    def generate(cls):

        _content = []
        for line in cls.raw.split("\n"):
            if len(line) < cls.w - 2:
                _content.append(line)
            else:
                _content.append(line[: cls.w - 2])
                _content.append(line[cls.w - 2 :])

        cls.content = _content
        cls.content_len = len(_content)  # record content length

    @classmethod
    def update(self):
        start_y = self.y + 1
        start_x = self.x + 1

        self.box_content = ""
        count_ = 0
        for line in self.content[self.content_selected_line :]:
            if isinstance(line, list):
                for sub_line in line:
                    if count_ < self.h - 2:
                        self.box_content += f"{Cursor.to(start_y, start_x)}{sub_line}"
                        start_y += 1
                        count_ += 1
                    else:
                        break
            else:
                if count_ < self.h - 2:
                    self.box_content += f"{Cursor.to(start_y, start_x)}{line}"
                    start_y += 1
                    count_ += 1

    @classmethod
    def notify(cls, notifier: Box):
        for sub in cls.__subclasses__():
            if notifier.genre & sub.genre:
                """
                if not sub.box:  # No box profile, create it.
                    sub.create_profile()
                """
                # Selected item has change.
                if sub.box_selected_idx != notifier.selected:
                    sub.content_selected_line = 0
                    sub.box_selected_idx = notifier.selected
                    sub.fetch_data(notifier)
                    sub.generate()
                sub.update()
                sub.render()
                break  # Finish the remaining cycle after the operation.

    @staticmethod
    def split_out_of_str(s: str, limit: int) -> list:
        idx = 0
        s_len = len(s)
        sub_lines = []
        while idx + limit < s_len:
            sub_lines.append(s[idx : idx + limit])
            idx += limit
        sub_lines.append(s[idx:])
        return sub_lines


# These subclasses will be called by the notify() method of the parent class
# to achieve the same type of content data for one layer cache. Subclasses
# override the generate() method of the parent class to color different types
# of content.
class StateContentBox(ContentBox):
    name: str = "Information"
    genre = BoxType.CONTENT | BoxType.STATE


class StatusContentBox(ContentBox):
    name: str = "File"
    genre = BoxType.CONTENT | BoxType.STATUS

    @classmethod
    def generate(cls):
        line_w = cls.w - 2
        _content = []
        lines_ = cls.raw.split("\n")
        lines_len = len(lines_)

        if lines_len < 4:
            super().generate()
            return

        # First four line be bolded.
        for i in range(4):
            _content.append(f"{Fx.b}{lines_[i]}{Fx.ub}")

        if lines_len > 5:
            if lines_[4].endswith("@@"):
                _content.append(f"{Theme.BOX_SELECTED_COLOR}{lines_[4]}{Theme.DEFAULT}")
            else:
                _content.append(lines_[4])

        if lines_len > 5:
            for line in lines_[5:]:
                line_len = len(line)
                if line_len <= line_w:
                    if line.startswith(ADDITION_FLAG):
                        line = f"{Theme.ADDITION}{line}{Theme.DEFAULT}"
                    elif line.startswith(DELETION_FLAG):
                        line = f"{Theme.DELETION}{line}{Theme.DEFAULT}"
                    _content.append(line)
                else:
                    sub_lines = cls.split_out_of_str(line, line_w)
                    if line.startswith(ADDITION_FLAG):
                        sub_lines[0] = Theme.ADDITION + sub_lines[0]
                        sub_lines[-1] = sub_lines[-1] + Theme.DEFAULT
                    elif line.startswith(DELETION_FLAG):
                        sub_lines[0] = Theme.DELETION + sub_lines[0]
                        sub_lines[-1] = sub_lines[-1] + Theme.DEFAULT
                    _content.append(sub_lines)

        cls.content = _content
        cls.content_len = len(_content)  # record content length


class BranchContentBox(ContentBox):
    name: str = "Log"
    genre = BoxType.CONTENT | BoxType.BRANCH


class CommitContentBox(ContentBox):
    name: str = "Patch"
    genre = BoxType.CONTENT | BoxType.COMMIT

    @classmethod
    def generate(cls):
        line_w = cls.w - 2
        _content = []
        lines_ = cls.raw.split("\n")
        lines_len = len(lines_)

        _content.append(f"{Theme.PUSHED}{Fx.b}{lines_[0]}{Fx.ub}{Theme.DEFAULT}")

        if lines_len >= 3:
            for i in range(1, 3):
                _content.append(f"{Fx.b}{lines_[i]}{Fx.ub}")

        if lines_len > 3:
            for line in lines_[3:]:
                line_len = len(line)
                if line_len <= line_w:
                    if line.startswith(ADDITION_FLAG):
                        line = f"{Theme.ADDITION}{line}{Theme.DEFAULT}"
                    elif line.startswith(DELETION_FLAG):
                        line = f"{Theme.DELETION}{line}{Theme.DEFAULT}"
                    _content.append(line)
                else:
                    sub_lines = cls.split_out_of_str(line, line_w)
                    if line.startswith(ADDITION_FLAG):
                        sub_lines[0] = Theme.ADDITION + sub_lines[0]
                        sub_lines[-1] = sub_lines[-1] + Theme.DEFAULT
                    elif line.startswith(DELETION_FLAG):
                        sub_lines[0] = Theme.DELETION + sub_lines[0]
                        sub_lines[-1] = sub_lines[-1] + Theme.DEFAULT
                    _content.append(sub_lines)

        cls.content = _content
        cls.content_len = len(_content)  # record content length


class StashContentBox(ContentBox):
    name: str = "Stash"
    genre = BoxType.CONTENT | BoxType.STASH


def fetch_content(c):
    selected = c.current

    if selected & BoxType.STATUS:
        _status_list = c.raw
        if not _status_list:
            return "No changed file."
        else:
            file_ = _status_list[c.selected]

            if not file_.tracked:  # is mean untrack
                return info.diff(file_.name, tracked=False)
            elif file_.has_staged_change:
                return info.diff(file_.name, cached=True)
            else:
                return info.diff(file_.name)
    elif selected & BoxType.COMMIT:
        args = c.raw
        if not args:
            return "No commits."
        else:
            _commit_id = args[c.selected].sha
            return info.commit_info(_commit_id)
    elif selected & BoxType.BRANCH:
        args = c.raw
        if not args:
            return "No local branches."
        else:
            _branch = args[c.selected]
            return info.branch_log(_branch)
    elif selected & BoxType.STASH:
        # TODO:
        return "Don't support display."
    elif selected & BoxType.STATE:
        return info.INTRODUCE


# _CACHED = re.compile(r'^[A-Z]\s$')



from ..style.color import Color


class Theme:
    DEFAULT = Color.fg("#ee")
    BOX_SELECTED_COLOR = Color.fg("#48D1CC")  # lightpink

    FILE_NEW = Color.fg("#4682B4")  # SteelBlue
    FILE_UNTRACK = Color.fg("#F08080")  # LightCoral

    FILE_CACHED = Color.fg("#87CEFA")  # LightSkyBlue
    FILE_RENAME = Color.fg("#87CEFA")
    FILE_DELED = Color.fg("#98FB98")

    FILE_DEL = Color.fg("#F08080")
    FILE_CHANGE = Color.fg("#F08080")

    BRANCH = Color.fg("#87CEFA")
    BRANCH_STATUS = Color.fg("#F0E68C")  # Linen

    PUSHED = Color.fg("#F0E68C")
    UNPUSHED = Color.fg("#F08080")

    ADDITION = Color.fg("#87CEFA")
    DELETION = Color.fg("#F08080")

    ERROR = Color.fg("#F08080")  # LightCoral



class File:
    def __init__(
        self,
        name: str,
        display_str: str,
        short_status: str,
        has_staged_change: bool,
        has_unstaged_change: bool,
        tracked: bool,
        deleted: bool,
        added: bool,
        has_merged_conflicts: bool,
        has_inline_merged_conflicts: bool,
    ) -> None:
        self.name = name
        self.display_str = display_str
        self.short_status = short_status
        self.has_staged_change = has_staged_change
        self.has_unstaged_change = has_unstaged_change
        self.tracked = tracked
        self.deleted = deleted
        self.added = added
        self.has_merged_conflicts = has_merged_conflicts
        self.has_inline_merged_conflicts = has_inline_merged_conflicts



import os
import logging

from . import Git, run_with_git, run_cmd


LOG = logging.getLogger(__name__)

# ==========================================================
# File option.
# ==========================================================
def stage(file_path: str) -> None:
    """Stage files."""

    command = "add --"
    if "->" in file_path:
        file_path = file_path.split("->")[-1].strip()

    run_with_git(" ".join([command, file_path]))


def stage_all() -> None:
    """Stage all files."""

    run_with_git("add -A")


def unstage(*args, tracked: bool = True) -> None:
    """Unstage files."""

    if tracked:
        command = "reset HEAD --"
    else:
        command = "rm --cached --force --"

    run_with_git(" ".join([command, *args]))


def unstage_all() -> None:
    """Unstage all files."""

    run_with_git("reset")


def pull() -> None:
    run_with_git("pull")


def push(force: bool = False, safe: bool = True):
    command_force = ""
    if force:
        if safe:
            command_force = "--force-with-lease"
        else:
            command_force = "--force"

    command = " ".join(["push", command_force])
    run_with_git(command)


def commit(message: str = ""):
    if message:
        pass
        return False
    else:
        run_cmd("git commit --verbose")
        return True


def del_(path: str, tracked: bool = True):
    if tracked:
        run_with_git(["checkout", path])
    else:
        os.remove(f"{Git.REPOSITORY_PATH}/{path}")


def ignore(path: str):
    with open(f"{Git.REPOSITORY_PATH}/.gitignore", "a") as f:
        f.write(f"\n{path}")


# ==========================================================
# Branch option.
# ==========================================================
def checkout(path: str):
    command = f"checkout {path}"
    err, resp = run_with_git(command)
    # LOG.debug(f"{err}|{resp}")
    return err, resp



from . import run_with_git


def branch_log(branch) -> str:
    """Gets all logs of a given branch.

    Args:
        branch: branch name
    """

    # if branch.startswith('* '):
    #     branch = branch[2:]
    branch_name = branch.name

    arg = "log %s --graph --decorate" % branch_name
    _, resp = run_with_git(arg)
    return resp.rstrip()


def commit_info(commit: str) -> str:
    """Gets the information for a given submission.

    Args:
        commit: commit id
    """

    return commit_file_info(commit)


def commit_file_info(commit: str, file_name: str = "") -> str:
    """Gets the change of a file in a given commit.

    Args:
        commit: commit id
        file_name: file name(include full path)
    """

    arg = "show %s %s" % (commit, file_name)
    _, resp = run_with_git(arg)
    return resp.rstrip()


def diff(file: str, tracked: bool = True, cached: bool = False) -> str:
    """Get the diff of given file.

    Args:
        file: file path
        tracked: Is the file tracked
        cached: Is the file staged
    """

    args = ["diff", "--submodule", "--no-ext-diff"]
    if cached:
        args.append("--cached")

    if not tracked:
        args.append("--no-index -- /dev/null")
    else:
        args.append("--")

    if "->" in file:  # rename
        file = file.split("->")[-1].strip()

    args.append(file)

    _, res = run_with_git(*args)
    return res.rstrip()


def is_selected_branch(branch: str) -> bool:
    """Determine whether a branch is the current branch."""

    return branch.startswith("* ")


# TODO: just temp introduce
INTRODUCE = """\
  __                   _ _
 / _|_   _ _ __   __ _(_) |_
| |_| | | | '_ \ / _` | | __|
|  _| |_| | | | | (_| | | |_
|_|  \__,_|_| |_|\__, |_|\__|
                 |___/

A terminal tool, help you use git more simple. Support Linux and MacOS.

"""



import subprocess
import os
import logging

LOG = logging.getLogger(__name__)


class Git(object):
    prefix = "git"
    REPOSITORY_PATH = ""


def run_with_git(*args) -> tuple:
    """Execute a git command and return the result.

    Args:
        args: Command parameter tuple

    Returns:
        Execution result, string text
    """

    command = " ".join([Git.prefix, *args])
    # LOG.debug(f"<run_with_git> {command}")
    try:
        with subprocess.Popen(
            [command], stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
        ) as proc:
            res = proc.stdout.read().decode()
            err = proc.stderr.read().decode()
            return err, res
    except Exception as e:
        LOG.warning(e)
        return e, ""


def run_cmd(*args):
    try:
        with subprocess.Popen(" ".join(args), shell=True) as proc:
            proc.wait()
    except Exception as e:
        LOG.warning(e)


def run_cmd_with_resp(*args) -> tuple:
    try:
        with subprocess.Popen(
            " ".join(args), stderr=subprocess.PIPE, stdout=subprocess.PIPE, shell=True
        ) as proc:
            res = proc.stdout.read().decode()
            err = proc.stderr.read().decode()
            return err, res
    except Exception as e:
        LOG.warning(e)
        print(e)
        return e, ""


def current_repository():
    err, path = run_with_git("rev-parse --git-dir")

    if err:
        return ""

    path = path.strip()
    if path == ".git":
        repository_path = os.getcwd()
    else:
        repository_path = path.split("/")[-2]
    return repository_path



import logging

from ..term import Term
from .key import Key


LOG = logging.getLogger(__name__)


def quit_app(code: int = 0, message: str = ""):
    Term.clear()
    Key.stop()
    if message:
        LOG.warning(message)
        print(message)
    raise SystemExit(code)



import threading
from typing import Dict


class Renderer:
    """Holds the draw buffer and manages IO blocking queue
    * .buffer([+]name[!], *args, append=False, now=False, z=100) : Add *args to buffer
    * - Adding "+" prefix to name sets append to True and appends to name's current string
    * - Adding "!" suffix to name sets now to True and print name's current string
    * .out(clear=False) : Print all strings in buffer, clear=True clear all buffers after
    * .now(*args) : Prints all arguments as a string
    * .clear(*names) : Clear named buffers, all if no argument
    """

    idle = threading.Event()
    idle.set()

    old_tree: Dict = dict()  # for diff
    _w: int = 0
    _h: int = 0
    _is_changed: bool = False

    @classmethod
    def now(cls, *args):
        """Wait for input reader and self to be idle then print to screen"""
        cls.idle.wait()
        cls.idle.clear()
        try:
            print(*args, sep="", end="", flush=True)
        except BlockingIOError:
            pass
            print(*args, sep="", end="", flush=True)
        cls.idle.set()



import time

from fungit.gui.manage import Manager
from .key import Key
from .clean_quit import quit_app


def process_key():
    while Key.has_event():
        key = Key.get()

        if key in ["mouse_scroll_up", "mouse_scroll_down", "mouse_click"]:
            mouse_pos = Key.get_mouse()
            Manager.mouse_event(key, mouse_pos)

        elif key == "q":
            quit_app()
        elif "1" <= key <= "5":
            Manager.switch_box_by_index(key)
        elif key == "h" or key == "left":
            Manager.prev_box()
        elif key == "l" or key == "right":
            Manager.next_box()
        elif key == "j" or key == "down":
            Manager.next_item()
        elif key == "k" or key == "up":
            Manager.prev_item()
        elif key == " ":
            Manager.space_event()
        elif key == "a":
            Manager.a_event()
        elif key == "p":
            Manager.pull()
        elif key == "P":
            Manager.push()
        elif key == "C":
            Manager.commit()
        elif key == "d":
            Manager.del_()
        elif key == "i":
            Manager.i_event()
        else:
            continue
    time.sleep(0.1)



class Cursor:
    """Class with collection of cursor movement functions: .t[o](line, column) | .r[ight](columns) | .l[eft](columns) | .u[p](lines) | .d[own](lines) | .save() | .restore()"""

    @staticmethod
    def to(line: int, col: int) -> str:
        # * Move cursor to line, column
        return "\033[{};{}f".format(line, col)

    @staticmethod
    def right(dx: int) -> str:
        return "\033[{}C".format(dx)

    @staticmethod
    def left(dx: int) -> str:
        return "\033[{}D".format(dx)

    @staticmethod
    def up(dy: int) -> str:
        return "\033[{}A".format(dy)

    @staticmethod
    def down(dy: int) -> str:
        return "\033[{}B".format(dy)

    save: str = "\033[s"  # * Save cursor position
    restore: str = "\033[u"  # * Restore saved cursor postion
    t = to
    r = right
    l = left
    u = up
    d = down



from typing import Tuple, Iterable


class Color:
    """Holds representations for a 24-bit color value
    __init__(color, depth="fg", default=False)
    -- color accepts 6 digit hexadecimal: string "#RRGGBB", 2 digit hexadecimal: string "#FF" or decimal RGB "255 255 255" as a string.
    -- depth accepts "fg" or "bg"
    __call__(*args) joins str arguments to a string and apply color
    __str__ returns escape sequence to set color
    __iter__ returns iteration over red, green and blue in integer values of 0-255.
    * Values:  .hexa: str  |  .dec: Tuple[int, int, int]  |  .red: int  |  .green: int  |  .blue: int  |  .depth: str  |  .escape: str
    """

    hexa: str
    dec: Tuple[int, int, int]
    red: int
    green: int
    blue: int
    depth: str
    escape: str
    default: bool

    TRUE_COLOR = False

    def __init__(self, color: str, depth: str = "fg", default: bool = False):
        self.depth = depth
        self.default = default
        try:
            if not color:
                self.dec = (-1, -1, -1)
                self.hexa = ""
                self.red = self.green = self.blue = -1
                self.escape = "\033[49m" if depth == "bg" and default else ""
                return

            elif color.startswith("#"):
                self.hexa = color
                if len(self.hexa) == 3:
                    self.hexa += self.hexa[1:3] + self.hexa[1:3]
                    c = int(self.hexa[1:3], base=16)
                    self.dec = (c, c, c)
                elif len(self.hexa) == 7:
                    self.dec = (
                        int(self.hexa[1:3], base=16),
                        int(self.hexa[3:5], base=16),
                        int(self.hexa[5:7], base=16),
                    )
                else:
                    raise ValueError(
                        f"Incorrectly formatted hexadecimal rgb string: {self.hexa}"
                    )

            else:
                c_t = tuple(map(int, color.split(" ")))
                if len(c_t) == 3:
                    self.dec = c_t  # type: ignore
                else:
                    raise ValueError('RGB dec should be "0-255 0-255 0-255"')

            ct = self.dec[0] + self.dec[1] + self.dec[2]
            if ct > 255 * 3 or ct < 0:
                raise ValueError(f"RGB values out of range: {color}")
        except Exception:
            # errlog.exception(str(e))
            self.escape = ""
            return

        if self.dec and not self.hexa:
            self.hexa = f'{hex(self.dec[0]).lstrip("0x").zfill(2)}{hex(self.dec[1]).lstrip("0x").zfill(2)}{hex(self.dec[2]).lstrip("0x").zfill(2)}'

        if self.dec and self.hexa:
            self.red, self.green, self.blue = self.dec
            self.escape = f'\033[{38 if self.depth == "fg" else 48};2;{";".join(str(c) for c in self.dec)}m'

        if Color.TRUE_COLOR:
            self.escape = f"{self.truecolor_to_256(rgb=self.dec, depth=self.depth)}"

    def __str__(self) -> str:
        return self.escape

    def __repr__(self) -> str:
        return repr(self.escape)

    def __iter__(self) -> Iterable:
        for c in self.dec:
            yield c

    # def __call__(self, *args: str) -> str:
    #     if len(args) < 1:
    #         return ""
    #     return f'{self.escape}{"".join(args)}{getattr(Term, self.depth)}'

    @staticmethod
    def truecolor_to_256(rgb: Tuple[int, int, int], depth: str = "fg") -> str:
        out: str = ""
        pre: str = f'\033[{"38" if depth == "fg" else "48"};5;'

        greyscale: Tuple[int, int, int] = (rgb[0] // 11, rgb[1] // 11, rgb[2] // 11)
        if greyscale[0] == greyscale[1] == greyscale[2]:
            out = f"{pre}{232 + greyscale[0]}m"
        else:
            out = f"{pre}{round(rgb[0] / 51) * 36 + round(rgb[1] / 51) * 6 + round(rgb[2] / 51) + 16}m"

        return out

    @staticmethod
    def escape_color(
        hexa: str = "", r: int = 0, g: int = 0, b: int = 0, depth: str = "fg"
    ) -> str:
        """Returns escape sequence to set color
        * accepts either 6 digit hexadecimal hexa="#RRGGBB", 2 digit hexadecimal: hexa="#FF"
        * or decimal RGB: r=0-255, g=0-255, b=0-255
        * depth="fg" or "bg"
        """
        dint: int = 38 if depth == "fg" else 48
        color: str = ""
        if hexa:
            try:
                if len(hexa) == 3:
                    c = int(hexa[1:], base=16)
                    if Color.TRUE_COLOR:
                        color = f"\033[{dint};2;{c};{c};{c}m"
                    else:
                        color = f"{Color.truecolor_to_256(rgb=(c, c, c), depth=depth)}"
                elif len(hexa) == 7:
                    if Color.TRUE_COLOR:
                        color = f"\033[{dint};2;{int(hexa[1:3], base=16)};{int(hexa[3:5], base=16)};{int(hexa[5:7], base=16)}m"
                    else:
                        color = f"{Color.truecolor_to_256(rgb=(int(hexa[1:3], base=16), int(hexa[3:5], base=16), int(hexa[5:7], base=16)), depth=depth)}"
            except ValueError:
                # errlog.exception(f'{e}')
                pass
        else:
            if Color.TRUE_COLOR:
                color = f"\033[{dint};2;{r};{g};{b}m"
            else:
                color = f"{Color.truecolor_to_256(rgb=(r, g, b), depth=depth)}"
        return color

    @classmethod
    def fg(cls, *args) -> str:
        if len(args) > 2:
            return cls.escape_color(r=args[0], g=args[1], b=args[2], depth="fg")
        else:
            return cls.escape_color(hexa=args[0], depth="fg")

    @classmethod
    def bg(cls, *args) -> str:
        if len(args) > 2:
            return cls.escape_color(r=args[0], g=args[1], b=args[2], depth="bg")
        else:
            return cls.escape_color(hexa=args[0], depth="bg")



class Commit:
    def __init__(
        self,
        sha: str,
        msg: str,
        author: str,
        unix_timestamp: int,
        status: str,
        extra_info: str,
        tag: list,
        action: str = "",
    ) -> None:
        self.sha = sha
        self.msg = msg
        self.author = author
        self.unix_timestamp = unix_timestamp
        self.status = status
        self.extra_info = extra_info
        self.tag = tag
        self.action = action

    def is_pushed(self):
        return self.status == "pushed"



import re
import logging
from typing import List

from . import run_with_git, Git
from .module.file import File
from .module.branch import Branch
from .module.commit import Commit


LOG = logging.getLogger(__name__)


def current_head():
    _, res = run_with_git("symbolic-ref -q --short HEAD")
    return res.rstrip()


def load_state() -> list:
    """Get current project name and head branch."""

    project_ = Git.REPOSITORY_PATH.split("/")[-1]
    head_ = current_head()
    return [project_, head_]


def load_files(*args) -> List[File]:
    """Get all file.

    Returns a list of `File` based on the current branch or head.

    Args:
        args: File tuple
    """

    command = "status -s -u"
    _, resp = run_with_git(" ".join([command, *args]))
    resp = resp.rstrip()

    files_ = []
    if resp:
        lines_ = resp.split("\n")
        for line_ in lines_:
            # _status.append((item[:2], item[3:]))
            change = line_[:2]
            staged_change = line_[:1]
            unstaged_change = line_[1:2]
            name = line_[3:]
            untracked = change in ["??", "A ", "AM"]
            has_no_staged_change = staged_change in [" ", "U", "?"]
            has_merged_conflicts = change in ["DD", "AA", "UU", "AU", "UA", "UD", "DU"]
            has_inline_merged_conflicts = change in ["UU", "AA"]

            file_ = File(
                name=name,
                display_str=line_,
                short_status=change,
                has_staged_change=not has_no_staged_change,
                has_unstaged_change=unstaged_change != " ",
                tracked=not untracked,
                deleted=unstaged_change == "D" or staged_change == "D",
                added=unstaged_change == "A" or untracked,
                has_merged_conflicts=has_merged_conflicts,
                has_inline_merged_conflicts=has_inline_merged_conflicts,
            )
            files_.append(file_)

    return files_


def load_branches() -> List[Branch]:
    command = 'for-each-ref --sort=-committerdate --format="%(HEAD)|%(refname:short)|%(upstream:short)|%(upstream:track)" refs/heads'
    err, resp = run_with_git(command)
    resp = resp.strip()

    if not resp:
        return []

    branchs = []
    lines = resp.split("\n")

    for line in lines:
        items = line.split("|")
        branch = Branch(
            name=items[1], pushables="?", pullables="?", is_head=items[0] == "*"
        )

        upstream_name = items[2]

        if not upstream_name:
            branchs.append(branch)
            continue

        branch.upstream_name = upstream_name

        track = items[3]
        _re = re.compile(r"ahead (\d+)")
        match = _re.search(track)
        if match:
            branch.pushables = str(match[1])
        else:
            branch.pushables = "0"

        _re = re.compile(r"behind (\d+)")
        match = _re.search(track)
        if match:
            branch.pullables = str(match[1])
        else:
            branch.pullables = "0"

        branchs.append(branch)

    return branchs


def get_first_pushed_commit(branch_name: str):
    command = "merge-base %s %s@{u}" % (branch_name, branch_name)
    _, resp = run_with_git(command)
    return resp.strip()


def get_log(branch_name: str, limit: bool = False, filter_path: str = ""):
    limit_flag = "-300" if limit else ""
    filter_flag = f"--follow -- {filter_path}" if filter_path else ""
    command = f'log {branch_name} --oneline --pretty=format:"%H|%at|%aN|%d|%p|%s" {limit_flag} --abbrev=20 --date=unix {filter_flag}'
    err, resp = run_with_git(command)
    return err, resp.strip()


def load_commits(branch_name: str):
    """Return given branch all commits.

    Args:
        branch_name: A string of branch name.
    """

    passed_first_pushed_commit = False
    first_pushed_commit = get_first_pushed_commit(branch_name)

    if not first_pushed_commit:
        passed_first_pushed_commit = True

    commits = []
    err, resp = get_log(branch_name)
    lines = resp.split("\n")
    if not err:
        for line in lines:
            split_ = line.split("|")

            sha = split_[0]
            unix_timestamp = int(split_[1])
            author = split_[2]
            extra_info = (split_[3]).strip()
            # parent_hashes = split_[4]
            message = "|".join(split_[5:])

            tag = []
            if extra_info:
                _re = re.compile(r"tag: ([^,\\]+)")
                match = _re.search(extra_info)
                if match:
                    tag.append(match[1])

            if sha == first_pushed_commit:
                passed_first_pushed_commit = True
            status = {True: "unpushed", False: "pushed"}[not passed_first_pushed_commit]

            commit_ = Commit(
                sha=sha,
                msg=message,
                author=author,
                unix_timestamp=unix_timestamp,
                status=status,
                extra_info=extra_info,
                tag=tag,
            )
            commits.append(commit_)

    return commits


def load_stashes() -> str:
    """Get stash list."""

    arg = "stash list"
    _, resp = run_with_git(arg)
    return resp



import os
import logging
import logging.handlers

from . import __FUNGITDIR__

FMT_NORMAL = logging.Formatter(
    fmt="%(asctime)s %(levelname).4s %(message)s", datefmt="%H:%M:%S"
)
FMT_DEBUG = logging.Formatter(
    fmt="%(asctime)s.%(msecs)03d %(levelname).4s [%(name)s] %(message)s",
    datefmt="%H:%M:%S",
)


def ensure_path(file_name: str) -> str:
    if not os.path.exists(file_name):
        dir_path = "/".join(file_name.split("/")[:-1])
        os.makedirs(dir_path, exist_ok=True)


def setup_logging(debug=False, log_file=__FUNGITDIR__ + "/log/fungit.log"):
    root_logger = logging.getLogger()

    if debug:
        log_level = logging.DEBUG
        formatter = FMT_DEBUG
    else:
        log_level = logging.INFO
        formatter = FMT_NORMAL

    if log_file:
        if log_file == "-":
            log_handle = logging.StreamHandler()
        else:
            ensure_path(log_file)
            try:
                log_handle = logging.handlers.RotatingFileHandler(
                    log_file, maxBytes=1048576, backupCount=4
                )
            except PermissionError:
                print(f'No permission to write to "{log_file}" directory!')
                raise SystemExit(1)

    log_handle.setFormatter(formatter)
    log_handle.setLevel(log_level)

    root_logger.addHandler(log_handle)
    root_logger.setLevel(0)



import logging
import threading

from .shared import BoxType, ConfirmType
from .box_option import generate_all_box as refresh_all
from .box.navigation_box import NavBox
from .box.git_box import GIT_BOXES
from .box.func_box import DynamicPromptBox, ConfirmBox
from .box.content_box import ContentBox
from fungit.commands import options


LOG = logging.getLogger(__name__)


class Manager:
    # box event
    @staticmethod
    def switch_box_by_index(idx):
        NavBox.set_current(GIT_BOXES[int(idx) - 1].genre)
        refresh_all(recreate=True, update_data=False)

    @staticmethod
    def prev_box():
        _current = NavBox.current
        _index = index_of(_current, need_box=False)
        _len = len(GIT_BOXES)
        new_index = _len - 1 - (_len - _index) % _len
        NavBox.set_current(GIT_BOXES[new_index].genre)
        refresh_all(recreate=True, update_data=False)

    @staticmethod
    def next_box():
        _current = NavBox.current
        _index = index_of(_current, need_box=False)
        _len = len(GIT_BOXES)
        new_index = (_index + 1) % _len
        NavBox.set_current(GIT_BOXES[new_index].genre)
        refresh_all(recreate=True, update_data=False)

    # item event
    @staticmethod
    def prev_item():
        _current = NavBox.current
        _, box = index_of(_current)

        if box.selected > 0:
            box.set_selected(box.selected - 1)

    @staticmethod
    def next_item():
        _current = NavBox.current
        _, box = index_of(_current)

        _max_len = len(box.raw) - 1

        if box.selected < _max_len:
            box.set_selected(box.selected + 1)

    @staticmethod
    def space_event():
        _current = NavBox.current
        LOG.debug(f"space event: {_current}")
        _, box = index_of(_current)

        if _current & BoxType.STATUS:
            file_ = box.raw[box.selected]

            if file_.has_unstaged_change:
                options.stage(file_.name)
            else:
                options.unstage(file_.name)
            box.notify(update_data=True)
        elif _current & BoxType.BRANCH:
            branch = box.raw[box.selected]
            err, _ = options.checkout(branch.name)
            if err and "error" in err:
                ConfirmBox.main("Error", err, ConfirmType.ERROR)
            refresh_all()
        elif _current & BoxType.COMMIT:
            commit = box.raw[box.selected]
            err, resp = options.checkout(commit.sha)
            if err and "error" in err:
                ConfirmBox.main("Error", err, ConfirmType.ERROR)
            refresh_all()
            LOG.debug(f"{err} | {resp}")
            # TODO: checkout commit has bug and error
        else:
            # TODO:
            pass

    @staticmethod
    def a_event():
        _current = NavBox.current
        _, box = index_of(_current)

        if _current & BoxType.STATUS:
            for file_ in box.raw:
                if file_.has_unstaged_change:
                    options.stage_all()
                    break
            else:
                options.unstage_all()

            box.notify(update_data=True)
        else:
            # TODO:
            pass

    @staticmethod
    def i_event():
        _current = NavBox.current
        _, box = index_of(_current)

        if _current == BoxType.STATUS:
            file = box.raw[box.selected]

            try:
                options.ignore(file.name)
            except Exception as e:
                ConfirmBox.main("Error", e, ConfirmType.ERROR)
            box.notify(update_data=True)
        else:
            pass

    @staticmethod
    def pull():
        def _pull_option():
            options.pull()
            DynamicPromptBox.close = True

        t = threading.Thread(target=_pull_option)
        t.setDaemon(True)
        t.start()

        DynamicPromptBox.main("", "Pulling... ")
        refresh_all()

    @staticmethod
    def push():
        def _push_option():
            options.push()
            DynamicPromptBox.close = True

        t = threading.Thread(target=_push_option)
        t.setDaemon(True)
        t.start()

        DynamicPromptBox.main("", "Pushing.. ")
        refresh_all()

    @staticmethod
    def commit():
        is_open_editor = options.commit()
        if is_open_editor:
            _current = NavBox.current
            _, box = index_of(_current)
            box.selected = 0
            refresh_all()

    @staticmethod
    def del_():
        _current = NavBox.current
        _, box = index_of(_current)

        if _current & BoxType.STATUS:
            is_confirm = ConfirmBox.main("", "Discard all changed?")
            if is_confirm:
                file = box.raw[box.selected]
                options.del_(file.name, file.tracked)
                refresh_all()
            else:
                refresh_all(update_data=False)

    @staticmethod
    def mouse_event(key, mouse_pos):
        _current = NavBox.current
        _, box = index_of(_current)
        sub = get_sub_content_box(box)
        mouse_x, mouse_y = mouse_pos
        LOG.debug(f"{key} {box} {sub} {sub.content_selected_line}")
        LOG.debug(f"{mouse_x} {mouse_y} {ContentBox.x} {ContentBox.y}")

        if mouse_x > ContentBox.x and mouse_y > ContentBox.y:
            LOG.debug(f"{ContentBox.h}")
            fresh = False
            if key == "mouse_scroll_down":
                if sub.content_selected_line < sub.content_len:
                    sub.content_selected_line += 1
                    fresh = True
            elif key == "mouse_scroll_up":
                if sub.content_selected_line > 0:
                    sub.content_selected_line -= 1
                    fresh = True
            if fresh:
                box.notify(update_data=False, re_profile=False)
        else:
            pass


def index_of(t, need_box: bool = True):
    for idx, sub in enumerate(GIT_BOXES):
        if sub.genre & t:
            if need_box:
                return idx, sub
            return idx


def get_sub_content_box(box):
    for sub in ContentBox.__subclasses__():
        if sub.genre & box.genre:
            return sub



import os
import math

from .box import Box
from .shared import BoxType, ConfirmType
from .box.navigation_box import NavBox
from .box.git_box import GIT_BOXES
from .box.content_box import ContentBox
from .box.func_box import ConfirmBox
from fungit.event.clean_quit import quit_app


def generate_all_box(
    recreate: bool = False, update_data: bool = True, lazy_render: bool = False
):
    if recreate:
        w, h = os.get_terminal_size()
        Box.t_w, Box.t_h = w, h
        generate_git_box_w_h()  # update all nav box (w)idth and (h)eigth.
        create_content_box()

    for sub in GIT_BOXES:
        if update_data:  # Update raw data.
            sub.fetch_data()
            sub.generate()
        sub.create_profile()
        sub.update()
        if not lazy_render:
            sub.render()


def generate_git_box_w_h():
    w, h = Box.t_w, Box.t_h

    _selected_type = NavBox.current
    limit_w = math.floor(w / 3)

    if w < 90 or h < 8:
        tip = "Not enough space!\nTerminal must be higher than 8 and wider than 90.\nPlease press `q` to exit."
        ConfirmBox.main("Error", tip, ConfirmType.ERROR, full=True)
        quit_app()

    elif h <= 20:
        _old = None
        for sub in GIT_BOXES:
            sub.x = 1
            sub.w = limit_w

            if not _old:
                sub.y = 1
            else:
                sub.y = _old.y + _old.h
            _old = sub

            if sub.genre & _selected_type:
                sub.h = h - (1 * 4) - 1
            else:
                sub.h = 1

    elif h <= 25:
        _old = None
        for sub in GIT_BOXES:
            sub.x = 1
            sub.w = limit_w

            if not _old:
                sub.y = 1
            else:
                sub.y = _old.y + _old.h
            _old = sub

            if sub.genre & _selected_type:
                sub.h = h - (3 * 4) - 1
            else:
                sub.h = 3

    else:
        temp_x = GIT_BOXES[0].x = 1
        temp_y = GIT_BOXES[0].y = 1
        temp_w = GIT_BOXES[0].w = limit_w
        temp_h = GIT_BOXES[0].h = 3

        if _selected_type & BoxType.STASH:
            _split_h, _less = divmod(h - 4, 4)
        else:
            _split_h, _less = divmod(h - 7, 3)

        for idx, sub in enumerate(GIT_BOXES[1:-1]):
            sub.x = 1
            sub.y = temp_y + temp_h
            sub.w = limit_w
            sub.h = _split_h + (1 if _less > idx else 0)

            temp_x, temp_y, temp_w, temp_h = sub.x, sub.y, sub.w, sub.h

        GIT_BOXES[-1].x = 1
        GIT_BOXES[-1].y = temp_y + temp_h
        GIT_BOXES[-1].w = limit_w
        if _selected_type & BoxType.STASH:
            GIT_BOXES[-1].h = _split_h
        else:
            GIT_BOXES[-1].h = 3


def create_content_box():
    w, h = Box.t_w, Box.t_h
    limit_w = math.floor(w / 3)

    ContentBox.x = limit_w + 1
    ContentBox.y = 1
    ContentBox.w = w - limit_w
    ContentBox.h = h - 1

    for sub_con in ContentBox.__subclasses__():
        sub_con.create_profile()



from fungit.gui.shared import BoxType
import os
import sys
import termios
import threading
from typing import List

from .gui.renderer import Renderer
from .commands import current_repository, Git


class Term:
    """Terminal info and commands"""

    width: int = 0
    height: int = 0
    resized: bool = False
    _w: int = 0
    _h: int = 0
    fg: str = ""  # * Default foreground color
    bg: str = ""  # * Default background color
    hide_cursor = "\033[?25l"  # * Hide terminal cursor
    show_cursor = "\033[?25h"  # * Show terminal cursor
    alt_screen = "\033[?1049h"  # * Switch to alternate screen
    normal_screen = "\033[?1049l"  # * Switch to normal screen
    clear_ = "\033[2J\033[0;0f"  # * Clear screen and set cursor to position 0,0
    # * Enable reporting of mouse position on click and release
    mouse_on = "\033[?1002h\033[?1015h\033[?1006h"
    mouse_off = "\033[?1002l"  # * Disable mouse reporting
    # * Enable reporting of mouse position at any movement
    mouse_direct_on = "\033[?1003h"
    mouse_direct_off = "\033[?1003l"  # * Disable direct mouse reporting
    winch = threading.Event()
    old_boxes: List = []
    min_width: int = 0
    min_height: int = 0

    @classmethod
    def init(cls):
        path_ = current_repository()
        if path_:
            Git.REPOSITORY_PATH = path_
        else:
            print("Error: please run it in git repository.")
            raise SystemExit(0)

        Term.width = os.get_terminal_size().columns
        Term.height = os.get_terminal_size().lines

        Renderer.now(
            Term.alt_screen,
            Term.clear_,
            Term.hide_cursor,
            Term.mouse_on,
            Term.title("fungit"),
        )
        cls.echo(False)

    @classmethod
    def clear(cls):
        Renderer.now(
            Term.clear_,
            Term.normal_screen,
            Term.show_cursor,
            Term.mouse_off,
            Term.mouse_direct_off,
            Term.title(),
        )
        cls.echo(True)

    @staticmethod
    def title(text: str = "") -> str:
        out: str = f'{os.environ.get("TERMINAL_TITLE", "")}'
        if out and text:
            out += " "
        if text:
            out += f"{text}"
        return f"\033]0;{out}\a"

    @staticmethod
    def echo(on: bool):
        """Toggle input echo"""
        (iflag, oflag, cflag, lflag, ispeed, ospeed, cc) = termios.tcgetattr(
            sys.stdin.fileno()
        )
        if on:
            lflag |= termios.ECHO  # type: ignore
        else:
            lflag &= ~termios.ECHO  # type: ignore
        new_attr = [iflag, oflag, cflag, lflag, ispeed, ospeed, cc]
        termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW, new_attr)



import logging

from fungit.style import Color, Fx
from . import __TERLOG__


LOG = logging.getLogger(__name__)


class CommandColor:
    """Terminal print color class."""

    RED = Color.fg("#FF6347")  # Tomato
    GREEN = Color.fg("#98FB98")  # PaleGreen
    YELLOW = Color.fg("#FFD700")  # Gold


def echo(msg: str, color="", style="", nl: bool = True):
    """Print to terminal.

    Print special information with color and style according to the
    incoming parameters.

    Args:
        msg: A special message.
        color: Message color.
        style: Message style, like: [bold, underline].
        nl: Is there a line feed.
    """
    print(f"{style}{color}{msg}{Fx.reset}", end="\n" if nl else "")


def okay(msg: str):
    """Print green information."""
    echo(f"{Fx.b}{CommandColor.GREEN}{msg}{Fx.reset}")


def warn(msg: str):
    """Print yellow information."""
    echo(f"{Fx.b}{CommandColor.YELLOW}{msg}{Fx.reset}")


def err(msg: str):
    """Print red information."""
    echo(f"{Fx.b}{CommandColor.RED}{msg}{Fx.reset}")


def exit_(*args):
    """Exit program.

    Receive error code, error message. If the error code matches, print the
    error information to the log. Then the command line output prompt, and
    finally exit.

    Args:
        *args:
            code: Exit code.
            msg: Error message.
    """
    if args and args[0] == 1:
        LOG.error(args[1:])
        print(f"Please check {__TERLOG__}")

    raise SystemExit(0)



import os
import time
import logging

from fungit.event.key import Key
from fungit.style import Cursor, Symbol
from fungit.event.clean_quit import quit_app
from ..utils import create_profile
from ..renderer import Renderer
from ..shared import ConfirmType
from ..theme import Theme
from . import Box


LOG = logging.getLogger(__name__)


class DynamicPromptBox(Box):

    close: bool = False
    count = 0
    process_symbol = [Symbol.graph_down[key] for key in Symbol.graph_down.keys()]

    @classmethod
    def main(cls, title, prompt):
        f_w, f_h = os.get_terminal_size()

        cls.x = round(f_w / 4)
        cls.w = round(f_w / 2)
        _w = cls.w - 2
        _len = len(prompt)
        idx = 0
        cls.content = []
        if _len > _w:
            while idx + _w < _len:
                cls.content.append(prompt[idx : idx + _w])
                idx += _w
        cls.content.append(prompt[idx:])
        cls.h = len(cls.content) + 2
        cls.y = round(f_h / 2) - round(cls.h / 2)

        cls.box = create_profile(cls.x, cls.y, cls.w, cls.h, title=title)
        Renderer.now(cls.box)

        start_x = cls.x + 1
        start_y = cls.y + 1
        line_w = cls.w - 2

        box_content = ""
        for idx, line in enumerate(cls.content):

            if idx < cls.h - 2:
                _line = f"{Cursor.to(start_y, start_x)}{line}"
                box_content += _line
                start_y += 1

        # cls.box_content += cls.process_symbol[cls.count % 4]
        # cls.count += 1
        cls.box_content = (
            box_content + cls.process_symbol[cls.count % len(cls.process_symbol)]
        )
        start_time = time.time()
        while not cls.close:
            if time.time() - start_time > 0.2:
                cls.count += 1
                cls.box_content = (
                    box_content
                    + cls.process_symbol[cls.count % len(cls.process_symbol)]
                )
                start_time = time.time()
            # cls.render()
            Renderer.now(cls.box_content)

            while Key.has_event():
                key = Key.get()

                if key == "q":
                    pass
                elif key == "scape":
                    cls.close = True
                    break
                else:
                    continue

        cls.close = False
        cls.count = 0


class ConfirmBox:
    close: bool = False

    @classmethod
    def main(
        cls, title: str, prompt: str, status=ConfirmType.NORMAL, full: bool = False
    ):
        f_w, f_h = os.get_terminal_size()

        # get box `x` and `(w)idth`
        if full:
            cls.x = 1
            cls.w = f_w
        else:
            cls.x = round(f_w / 4)
            cls.w = round(f_w / 2)

        line_len = cls.w - 2
        prompt = prompt.replace("\t", "")
        prompts = prompt.split("\n")

        LOG.debug(prompts)

        # process prompt string.
        cls.content = []
        for line in prompts:
            prompt_len = len(line)
            idx: int = 0
            if prompt_len > line_len:
                while idx + line_len < prompt_len:
                    cls.content.append(line[idx : idx + line_len])
                    idx += line_len
            cls.content.append(line[idx:])

        # get box `y` and `(h)eigth`.
        if full:
            cls.h = f_h
            cls.y = 1
        else:
            cls.h = len(cls.content) + 2
            cls.y = round(f_h / 2) - round(cls.h / 2)

        # get color
        color = cls.status_color(status)

        # create box profile
        cls.box = create_profile(
            cls.x, cls.y, cls.w, cls.h, title=title, line_color=color
        )

        # create the content in box.
        start_x = cls.x + 1
        start_y = cls.y + 1

        cls.box_content = ""
        for idx, line in enumerate(cls.content):

            if idx < cls.h - 2:
                _line = f"{Cursor.to(start_y, start_x)}{color}{line}{Theme.DEFAULT}"
                cls.box_content += _line
                start_y += 1

        # render
        Renderer.now(cls.box, cls.box_content)

        # listen key
        is_confirm: bool = False
        Key.clear()
        while not cls.close:
            while Key.has_event():
                key = Key.get()
                LOG.debug(key)

                if key == "q":
                    cls.close = True
                    quit_app()
                elif key in ["enter", "y", "Y", " "]:
                    cls.close = True
                    is_confirm = True
                    break
                elif key == "scape":  # escape
                    cls.close = True
                    break
                else:
                    continue
            time.sleep(0.1)

        Key.clear()
        cls.close = False
        return is_confirm

    @classmethod
    def status_color(cls, status):
        if status & ConfirmType.ERROR:
            return Theme.ERROR
        else:
            return ""


class InputBox:
    pass



import re

from fungit.commands import run_cmd
from .shared import warn


class GitOptionState:
    NO = 0
    ONE = 1
    MULTI = 1 << 1
    INTERC = 1 << 2
    FUNC = 1 << 3
    STRING = 1 << 4


def add(args: list):
    if args:
        args_str = " ".join(args)
    else:
        args_str = " ."

    run_cmd("git add " + args_str)


def fetch_remote_branch(args: list):
    branch = args[0] if len(args) > 1 else None

    if branch:
        run_cmd(f"git fetch origin {branch}:{branch} ")
    else:
        warn("This option need a branch name.")


def set_email_and_username(args: list):
    __global = re.compile(r"\-\-global")
    res = []
    for i in args:
        r = __global.search(i)
        if r is not None:
            res.append(i)
    if res:
        other = " --global "
    else:
        other = " "

    name = input("Please input username:")
    run_cmd(GIT_OPTIONS["user"]["command"] + other + name)
    email = input("Please input email:")
    run_cmd(GIT_OPTIONS["email"]["command"] + other + email)


def process_func(c: str, args: list):
    fn = GIT_OPTIONS[c]["command"]
    fn(args)


def process_origin_command(c: str, args: list):
    origin_command = GIT_OPTIONS[c]["command"]

    if args:
        args_str = " ".join(args)
        command = " ".join([origin_command, args_str])
    else:
        command = origin_command

    warn(command)
    run_cmd(command)


def process(c: str, args: list = None):
    try:
        state = GIT_OPTIONS[c]["state"]
    except Exception:
        print("Don't support this command, please try ", end="")
        warn("g --show-commands")
        raise SystemExit(0)

    if state & GitOptionState.FUNC:
        process_func(c, args)
    elif state & GitOptionState.STRING:
        process_origin_command(c, args)


TYPES = [
    "Branch",
    "Commit",
    "Conflict",
    "Fetch",
    "Index",
    "Log",
    "Merge",
    "Push",
    "Remote",
    "Stash",
    "Tag",
    "Working tree",
    "Setting",
]

# Global command dictionary, including attributes:
#     command type,
#     complete command,
#     help information
GIT_OPTIONS = {
    # Branch
    "b": {
        "state": GitOptionState.STRING | GitOptionState.NO,
        "command": "git branch ",
        "help-msg": "lists, creates, renames, and deletes branches.",
    },
    "bc": {
        "state": GitOptionState.STRING | GitOptionState.ONE,
        "command": "git checkout -b ",
        "help-msg": "creates a new branch.",
    },
    "bl": {
        "state": GitOptionState.STRING | GitOptionState.NO,
        "command": "git branch -vv ",
        "help-msg": "lists branches and their commits.",
    },
    "bL": {
        "state": GitOptionState.STRING | GitOptionState.NO,
        "command": "git branch --all -vv ",
        "help-msg": "lists local and remote branches and their commits.",
    },
    "bs": {
        "state": GitOptionState.STRING | GitOptionState.NO,
        "command": "git show-branch ",
        "help-msg": "lists branches and their commits with ancestry graphs.",
    },
    "bS": {
        "state": GitOptionState.STRING | GitOptionState.NO,
        "command": "git show-branch --all ",
        "help-msg": "lists local and remote branches and their commits with ancestry graphs.",
    },
    "bm": {
        "state": GitOptionState.STRING | GitOptionState.ONE,
        "command": "git branch --move ",
        "help-msg": "renames a branch.",
    },
    "bM": {
        "state": GitOptionState.STRING | GitOptionState.ONE,
        "command": "git branch --move --force ",
        "help-msg": "renames a branch even if the new branch name already exists.",
    },
    # Commit
    "c": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git commit --verbose ",
        "help-msg": "records changes to the repository.",
    },
    "ca": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git commit --verbose --all ",
        "help-msg": "commits all modified and deleted files.",
    },
    "cA": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git commit --verbose --patch ",
        "help-msg": "commits all modified and deleted files interactivly.",
    },
    "cm": {
        "state": GitOptionState.STRING | GitOptionState.ONE,
        "command": "git commit --verbose --message ",
        "help-msg": "commits with the given message.",
    },
    "co": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git checkout ",
        "help-msg": "checks out a branch or paths to the working tree.",
    },
    "cO": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git checkout --pathc ",
        "help-msg": "checks out hunks from the index or the tree interactively.",
    },
    "cf": {
        "state": GitOptionState.STRING | GitOptionState.NO,
        "command": "git commit --amend --reuse-message HEAD ",
        "help-msg": "amends the tip of the current branch reusing the same log message as HEAD.",
    },
    "cF": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git commit --verbose --amend ",
        "help-msg": "amends the tip of the current branch.",
    },
    "cr": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git revert ",
        "help-msg": "reverts existing commits by reverting patches and recording new commits.",
    },
    "cR": {
        "state": GitOptionState.STRING | GitOptionState.NO,
        "command": 'git reset "HEAD^" ',
        "help-msg": "removes the HEAD commit.",
    },
    "cs": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": 'git show --pretty=format:"%C(bold yellow)commit %H%C(auto)%d%n%C(bold)Author: %C(blue)%an <%ae> %C(reset)%C(cyan)%ai (%ar)%n%C(bold)Commit: %C(blue)%cn <%ce> %C(reset)%C(cyan)%ci (%cr)%C(reset)%n%+B"',
        "help-msg": "shows one or more objects (blobs, trees, tags and commits).",
    },
    # Conflict(C)
    "Cl": {
        "state": GitOptionState.STRING | GitOptionState.NO,
        "command": "git --no-pager diff --diff-filter=U --name-only ",
        "help-msg": "lists unmerged files.",
    },
    "Ca": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git add git --no-pager diff --diff-filter=U --name-only ",
        "help-msg": "adds unmerged file contents to the index.",
    },
    "Ce": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git mergetool git --no-pager diff --diff-filter=U --name-only ",
        "help-msg": "executes merge-tool on all unmerged files.",
    },
    "Co": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git checkout --ours -- ",
        "help-msg": "checks out our changes for unmerged paths.",
    },
    "CO": {
        "state": GitOptionState.STRING | GitOptionState.NO,
        "command": "git checkout --ours -- git --no-pager diff --diff-filter=U --name-only ",
        "help-msg": "checks out our changes for all unmerged paths.",
    },
    "Ct": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git checkout --theirs -- ",
        "help-msg": "checks out their changes for unmerged paths.",
    },
    "CT": {
        "state": GitOptionState.STRING | GitOptionState.NO,
        "command": "git checkout --theirs -- git --no-pager diff --diff-filter=U --name-only ",
        "help-msg": "checks out their changes for all unmerged paths.",
    },
    # Fetch(f)
    "f": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git fetch ",
        "help-msg": "downloads objects and references from another repository.",
    },
    "fc": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git clone ",
        "help-msg": "clones a repository into a new directory.",
    },
    "fC": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git clone --depth=1 ",
        "help-msg": "clones a repository into a new directory clearly(depth:1).",
    },
    "fm": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git pull ",
        "help-msg": "fetches from and merges with another repository or local branch.",
    },
    "fr": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git pull --rebase ",
        "help-msg": "fetches from and rebase on top of another repository or local branch.",
    },
    "fu": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git fetch --all --prune && git merge --ff-only @{u} ",
        "help-msg": "removes un-existing remote-tracking references, fetches all remotes and merges.",
    },
    "fb": {
        "state": GitOptionState.FUNC | GitOptionState.ONE,
        "command": fetch_remote_branch,
        "help-msg": "fetch other branch to local as same name.",
    },
    # Index(i)
    "ia": {
        "state": GitOptionState.FUNC | GitOptionState.MULTI,
        "command": add,
        "help-msg": "adds file contents to the index(default: all files).",
    },
    "iA": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git add --patch ",
        "help-msg": "adds file contents to the index interactively.",
    },
    "iu": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git add --update ",
        "help-msg": "adds file contents to the index (updates only known files).",
    },
    "id": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git diff --no-ext-diff --cached ",
        "help-msg": "displays changes between the index and a named commit (diff).",
    },
    "iD": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git diff --no-ext-diff --cached --word-diff ",
        "help-msg": "displays changes between the index and a named commit (word diff).",
    },
    "ir": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git reset ",
        "help-msg": "resets the current HEAD to the specified state.",
    },
    "iR": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git reset --patch ",
        "help-msg": "resets the current index interactively.",
    },
    "ix": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git rm --cached -r ",
        "help-msg": "removes files from the index (recursively).",
    },
    "iX": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git rm --cached -rf ",
        "help-msg": "removes files from the index (recursively and forced).",
    },
    # Log(l)
    "l": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git log --graph --all --decorate ",
        "help-msg": "displays the log with good format.",
    },
    "l1": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git log --graph --all --decorate --oneline ",
        "help-msg": "",
    },
    "ls": {
        "state": GitOptionState.STRING | GitOptionState.NO,
        "command": 'git log --topo-order --stat --pretty=format:"%C(bold yellow)commit %H%C(auto)%d%n%C(bold)Author: %C(blue)%an <%ae> %C(reset)%C(cyan)%ai (%ar)%n%C(bold)Commit: %C(blue)%cn <%ce> %C(reset)%C(cyan)%ci (%cr)%C(reset)%n%+B" ',
        "help-msg": "displays the stats log.",
    },
    "ld": {
        "state": GitOptionState.STRING | GitOptionState.NO,
        "command": 'git log --topo-order --stat --patch --pretty=format:"%C(bold yellow)commit %H%C(auto)%d%n%C(bold)Author: %C(blue)%an <%ae> %C(reset)%C(cyan)%ai (%ar)%n%C(bold)Commit: %C(blue)%cn <%ce> %C(reset)%C(cyan)%ci (%cr)%C(reset)%n%+B" ',
        "help-msg": "displays the diff log.",
    },
    "lv": {
        "state": GitOptionState.STRING | GitOptionState.NO,
        "command": 'git log --topo-order --show-signature --pretty=format:"%C(bold yellow)commit %H%C(auto)%d%n%C(bold)Author: %C(blue)%an <%ae> %C(reset)%C(cyan)%ai (%ar)%n%C(bold)Commit: %C(blue)%cn <%ce> %C(reset)%C(cyan)%ci (%cr)%C(reset)%n%+B" ',
        "help-msg": "displays the log, verifying the GPG signature of commits.",
    },
    "lc": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git shortlog --summary --numbered ",
        "help-msg": "displays the commit count for each contributor in descending order.",
    },
    "lr": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git reflog ",
        "help-msg": "manages reflog information.",
    },
    # Merge(m)
    "m": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git merge ",
        "help-msg": "joins two or more development histories together.",
    },
    "ma": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git merge --abort ",
        "help-msg": "aborts the conflict resolution, and reconstructs the pre-merge state.",
    },
    "mC": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git merge --no-commit ",
        "help-msg": "performs the merge but does not commit.",
    },
    "mF": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git merge --no-ff ",
        "help-msg": "creates a merge commit even if the merge could be resolved as a fast-forward.",
    },
    "mS": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git merge -S ",
        "help-msg": "performs the merge and GPG-signs the resulting commit.",
    },
    "mv": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git merge --verify-signatures ",
        "help-msg": "verifies the GPG signature of the tip commit of the side branch being merged.",
    },
    "mt": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git mergetool ",
        "help-msg": "runs the merge conflict resolution tools to resolve conflicts.",
    },
    # Push(p)
    "p": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git push ",
        "help-msg": "updates remote refs along with associated objects.",
    },
    "pf": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git push --force-with-lease ",
        "help-msg": 'forces a push safely (with "lease").',
    },
    "pF": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git push --force ",
        "help-msg": "forces a push. ",
    },
    "pa": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git push --all ",
        "help-msg": "pushes all branches.",
    },
    "pA": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git push --all && git push --tags ",
        "help-msg": "pushes all branches and tags.",
    },
    "pt": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git push --tags ",
        "help-msg": "pushes all tags.",
    },
    "pc": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": 'git push --set-upstream origin "$(git symbolic-ref -q --short HEAD 2> /dev/null)" ',
        "help-msg": "pushes the current branch and adds origin as an upstream reference for it.",
    },
    "pp": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": 'git pull origin "$(git symbolic-ref -q --short HEAD 2> /dev/null)" && git push origin "$(git symbolic-ref -q --short HEAD 2> /dev/null)" ',
        "help-msg": "pulls and pushes the current branch from origin to origin.",
    },
    # Remote(R)
    "R": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git remote ",
        "help-msg": "manages tracked repositories.",
    },
    "Rl": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git remote --verbose ",
        "help-msg": "lists remote names and their URLs.",
    },
    "Ra": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git remote add ",
        "help-msg": "adds a new remote.",
    },
    "Rx": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git remote rm ",
        "help-msg": "removes a remote.",
    },
    "Rm": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git remote rename ",
        "help-msg": "renames a remote.",
    },
    "Ru": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git remote update ",
        "help-msg": "fetches remotes updates.",
    },
    "Rp": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git remote prune ",
        "help-msg": "prunes all stale remote tracking branches.",
    },
    "Rs": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git remote show ",
        "help-msg": "shows information about a given remote.",
    },
    "RS": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git remote set-url ",
        "help-msg": "changes URLs for a remote.",
    },
    # Stash(s)
    "s": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git stash ",
        "help-msg": "stashes the changes of the dirty working directory.",
    },
    "sp": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git stash pop ",
        "help-msg": "removes and applies a single stashed state from the stash list.",
    },
    "sl": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git stash list ",
        "help-msg": "lists stashed states.",
    },
    "sd": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git stash show",
        "help-msg": "",
    },
    "sD": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git stash show --patch --stat",
        "help-msg": "",
    },
    # 'sr': {
    #     'state': GitOptionState.STRING | GitOptionState.MULTI,
    #     'command': '_git_stash_recover ',
    #     'help-msg': '',
    # },
    # 'sc': {
    #     'state': GitOptionState.STRING | GitOptionState.MULTI,
    #     'command': '_git_clear_stash_interactive',
    #     'help-msg': '',
    # },
    # Tag (t)
    "t": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git tag ",
        "help-msg": "creates, lists, deletes or verifies a tag object signed with GPG.",
    },
    "ta": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git tag -a ",
        "help-msg": "create a new tag.",
    },
    "tx": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git tag --delete ",
        "help-msg": "deletes tags with given names.",
    },
    # Working tree(w)
    "ws": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git status --short ",
        "help-msg": "displays working-tree status in the short format.",
    },
    "wS": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git status ",
        "help-msg": "displays working-tree status.",
    },
    "wd": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git diff --no-ext-diff ",
        "help-msg": "displays changes between the working tree and the index (diff).",
    },
    "wD": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git diff --no-ext-diff --word-diff ",
        "help-msg": "displays changes between the working tree and the index (word diff).",
    },
    "wr": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git reset --soft ",
        "help-msg": "resets the current HEAD to the specified state, does not touch the index nor the working tree.",
    },
    "wR": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git reset --hard ",
        "help-msg": "resets the current HEAD, index and working tree to the specified state.",
    },
    "wc": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git clean --dry-run ",
        "help-msg": "cleans untracked files from the working tree (dry-run).",
    },
    "wC": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git clean -d --force ",
        "help-msg": "cleans untracked files from the working tree.",
    },
    "wm": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git mv ",
        "help-msg": "moves or renames files.",
    },
    "wM": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git mv -f ",
        "help-msg": "moves or renames files (forced).",
    },
    "wx": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git rm -r ",
        "help-msg": "removes files from the working tree and from the index (recursively).",
    },
    "wX": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git rm -rf ",
        "help-msg": "removes files from the working tree and from the index (recursively and forced).",
    },
    # Setting
    "savepd": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git config credential.helper store ",
        "help-msg": "Remember your account and password.",
    },
    "ue": {
        "state": GitOptionState.FUNC | GitOptionState.NO,
        "command": set_email_and_username,
        "help-msg": "set email and username interactively.",
    },
    "user": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git config user.name ",
        "help-msg": "",
    },
    "email": {
        "state": GitOptionState.STRING | GitOptionState.MULTI,
        "command": "git config user.email ",
        "help-msg": "",
    },
    # 'clear': {
    #     'state': GitOptionState.STRING | GitOptionState.MULTI,
    #     'command': '_git_clear ${@:2:$((${#@}))} ',
    #     'help-msg': '',
    # },
    # 'ignore': {
    #     'state': GitOptionState.STRING | GitOptionState.MULTI,
    #     'command': '_git_ignore_files ${@:2:$((${#@}))} ',
    #     'help-msg': '',
    # },
}


def is_branch_option(x):
    return x.startswith("b")


def is_commit_option(x):
    return x.startswith("c")


def is_conflict_option(x):
    return x.startswith("C")


def is_fetch_option(x):
    return x.startswith("f")


def is_index_option(x):
    return x.startswith("i")


def is_log_option(x):
    return x.startswith("l")


def is_merge_option(x):
    return x.startswith("m")


def is_push_option(x):
    return x.startswith("p")


def is_remote_option(x):
    return x.startswith("R")


def is_stash_option(x):
    return x.startswith("s")


def is_tag_option(x):
    return x.startswith("t")


def is_tree_option(x):
    return x.startswith("w")


def detect_option_type(k: str) -> str:
    pass



import logging
from typing import List, Any

from fungit.commands import loading
from fungit.style import Fx, Cursor, Symbol
from ..shared import BoxType
from ..theme import Theme
from .navigation_box import NavBox


LOG = logging.getLogger(__name__)


class StateBox(NavBox):
    name: str = "state"
    genre = BoxType.STATE
    x: int = 0
    y: int = 0
    w: int = 0
    h: int = 0
    raw: Any = None
    content: List = []
    selected: int = 0
    box: str = ""
    box_content: str = ""

    @classmethod
    def fetch_data(cls):
        cls.raw = loading.load_state()

    @classmethod
    def generate(cls):
        _project, _head = cls.raw
        cls.content = [f"{_project} {Symbol.right} {_head}"]


class StatusBox(NavBox):
    name: str = "status"
    genre = BoxType.STATUS
    x: int = 0
    y: int = 0
    w: int = 0
    h: int = 0
    raw: Any = None
    content: List = []
    selected: int = 0
    box: str = ""
    box_content: str = ""

    @classmethod
    def fetch_data(cls):
        cls.raw = loading.load_files()
        try:
            cls.raw[cls.selected]
        except IndexError:  # out of range (rename, ignore)
            cls.selected -= 1

    @classmethod
    def generate(cls):
        line_w = cls.w - 2
        _content = []

        for file_ in cls.raw:
            str_len = len(file_.display_str)
            display_str = (
                file_.display_str if str_len > line_w else file_.display_str[:line_w]
            )
            color_ = cls.line_color(file_)
            line_ = f"{color_}{display_str}{Theme.DEFAULT}"
            _content.append(line_)

        cls.content = _content

    @classmethod
    def update(cls):
        start_x = cls.x + 1
        start_y = cls.y + 1

        _current = cls.selected
        _limit = 0
        if _current + 1 >= cls.h - 2:  # current selected index big than height
            _limit = _current + 1 - (cls.h - 2)

        cls.box_content = ""
        for idx, line in enumerate(cls.content):
            if idx >= _limit and idx - _limit < cls.h - 2:
                _line = f"{Cursor.to(start_y, start_x)}{line}"
                if cls.genre & cls.current and idx == _current:
                    _line = f"{Fx.b}{_line}{Fx.ub}"
                cls.box_content += _line
                start_y += 1

    @classmethod
    def line_color(cls, f):
        color = Theme.DEFAULT
        if f.tracked:
            if f.has_unstaged_change:
                color = Theme.FILE_CHANGE
            elif f.has_staged_change:
                color = Theme.FILE_CACHED
            elif f.deleted:
                color = Theme.FILE_DEL
        else:  # untracked
            if f.has_unstaged_change:
                color = Theme.FILE_UNTRACK
            else:
                color = Theme.FILE_NEW

        return color


class BranchBox(NavBox):
    name: str = "branch"
    genre = BoxType.BRANCH
    x: int = 0
    y: int = 0
    w: int = 0
    h: int = 0
    raw: Any = None
    content: List = []
    selected: int = 0
    box: str = ""
    box_content: str = ""

    @classmethod
    def fetch_data(cls):
        cls.raw = loading.load_branches()

    @classmethod
    def generate(cls):
        line_w = cls.w - 2
        content_ = []

        for branch in cls.raw:
            prefix_ = "* " if branch.is_head else "  "
            prefix_len = len(prefix_)

            status_, status_len = "", 0
            if branch.upstream_name:
                status_ = (
                    f" {Symbol.up}{branch.pushables}{Symbol.down}{branch.pullables}"
                )
                status_len = len(branch.pushables) + len(branch.pullables) + 3

            less_len = line_w - prefix_len - status_len
            name_ = branch.name
            name_len = len(branch.name)
            if name_len > less_len:
                name_ = f"{name_[:less_len - 3]}..."

            line_ = f"{Theme.BRANCH}{prefix_}{Theme.DEFAULT}{name_}{Theme.BRANCH_STATUS}{status_}{Theme.DEFAULT}"
            content_.append(line_)

        cls.content = content_

    @classmethod
    def update(cls):
        start_x = cls.x + 1
        start_y = cls.y + 1

        _current = cls.selected
        _limit = 0
        if _current + 1 >= cls.h - 2:  # current selected index big than heigth
            _limit = _current + 1 - (cls.h - 2)

        cls.box_content = ""
        for idx, line in enumerate(cls.content):
            if idx >= _limit and idx - _limit < cls.h - 2:
                _line = f"{Cursor.to(start_y, start_x)}{line}"
                if cls.genre & cls.current and idx == _current:
                    _line = f"{Fx.b}{_line}{Fx.ub}"
                cls.box_content += _line
                start_y += 1


class CommitBox(NavBox):
    name: str = "commit"
    genre = BoxType.COMMIT
    x: int = 0
    y: int = 0
    w: int = 0
    h: int = 0
    raw: Any = None
    content: List = []
    selected: int = 0
    box: str = ""
    box_content: str = ""

    @classmethod
    def fetch_data(cls):
        # get current head commit list
        cls.raw = loading.load_commits(loading.current_head())

    @classmethod
    def generate(cls):
        line_w = cls.w - 2 - 9
        content_ = []

        for commit in cls.raw:
            color_ = Theme.PUSHED if commit.is_pushed() else Theme.UNPUSHED
            id_ = commit.sha[:7]
            msg_ = commit.msg
            msg_ = msg_ if len(msg_) <= line_w else msg_[:line_w]
            content_.append(f"{color_}{id_} {Theme.DEFAULT}{msg_}")

        cls.content = content_

    @classmethod
    def update(cls):
        start_x = cls.x + 1
        start_y = cls.y + 1

        _current = cls.selected
        _limit = 0
        if _current + 1 >= cls.h - 2:  # current selected index big than height
            _limit = _current + 1 - (cls.h - 2)

        cls.box_content = ""
        for idx, line in enumerate(cls.content):

            if idx >= _limit and idx - _limit < cls.h - 2:
                _line = f"{Cursor.to(start_y, start_x)}{line}"
                if cls.genre & cls.current and idx == _current:
                    _line = f"{Fx.b}{_line}{Fx.ub}"
                cls.box_content += _line
                start_y += 1


class StashBox(NavBox):
    name: str = "stash"
    genre = BoxType.STASH
    x: int = 0
    y: int = 0
    w: int = 0
    h: int = 0
    raw: Any = None
    content: List = []
    selected: int = 0
    box: str = ""
    box_content: str = ""

    @classmethod
    def fetch_data(cls):
        cls.raw = loading.load_stashes()

    @classmethod
    def generate(cls):

        if cls.raw:
            cls.content = cls.raw.split("\n")
        else:
            cls.content = []


GIT_BOXES = [StateBox, StatusBox, BranchBox, CommitBox, StashBox]



class Branch:
    def __init__(
        self,
        name: str,
        pushables: str,
        pullables: str,
        is_head: bool,
        upstream_name: str = "",
    ):
        self.name = name
        self.upstream_name = upstream_name
        self.pushables = pushables
        self.pullables = pullables
        self.is_head = is_head



import threading
import tty
import termios
import fcntl
import os
import sys
import logging
from time import sleep, time
from select import select
from typing import List, Dict, Tuple, Union


LOG = logging.getLogger(__name__)


class Timer:
    timestamp: float
    return_zero = False

    @classmethod
    def stamp(cls):
        cls.timestamp = time()

    @classmethod
    def not_zero(cls) -> bool:
        if cls.return_zero:
            cls.return_zero = False
            return False
        return cls.timestamp + (1) > time()

    @classmethod
    def left(cls) -> float:
        return cls.timestamp + (1) - time()

    @classmethod
    def finish(cls):
        cls.return_zero = True
        cls.timestamp = time() - (1)
        Key.break_wait()


class Raw(object):
    """Set raw input mode for device"""

    def __init__(self, stream):
        self.stream = stream
        self.fd = self.stream.fileno()

    def __enter__(self):
        self.original_stty = termios.tcgetattr(self.stream)
        tty.setcbreak(self.stream)

    def __exit__(self, type, value, traceback):
        termios.tcsetattr(self.stream, termios.TCSANOW, self.original_stty)


class Nonblocking(object):
    """Set nonblocking mode for device"""

    def __init__(self, stream):
        self.stream = stream
        self.fd = self.stream.fileno()

    def __enter__(self):
        self.orig_fl = fcntl.fcntl(self.fd, fcntl.F_GETFL)
        fcntl.fcntl(self.fd, fcntl.F_SETFL, self.orig_fl | os.O_NONBLOCK)

    def __exit__(self, *args):
        fcntl.fcntl(self.fd, fcntl.F_SETFL, self.orig_fl)


class Key:
    """Handles the threaded input reader for keypresses and mouse events"""

    list: List[str] = []
    mouse: Dict[str, List[List[int]]] = {}
    mouse_pos: Tuple[int, int] = (0, 0)
    escape: Dict[Union[str, Tuple[str, str]], str] = {
        "\n": "enter",
        ("\x7f", "\x08"): "backspace",
        ("[A", "OA"): "up",
        ("[B", "OB"): "down",
        ("[D", "OD"): "left",
        ("[C", "OC"): "right",
        "[2~": "insert",
        "[3~": "delete",
        "[H": "home",
        "[F": "end",
        "[5~": "page_up",
        "[6~": "page_down",
        "\t": "tab",
        "[Z": "shift_tab",
        "OP": "f1",
        "OQ": "f2",
        "OR": "f3",
        "OS": "f4",
        "[15": "f5",
        "[17": "f6",
        "[18": "f7",
        "[19": "f8",
        "[20": "f9",
        "[21": "f10",
        "[23": "f11",
        "[24": "f12",
    }
    new = threading.Event()
    idle = threading.Event()
    mouse_move = threading.Event()
    mouse_report: bool = False
    idle.set()
    stopping: bool = False
    started: bool = False
    reader: threading.Thread

    @classmethod
    def start(cls):
        cls.stopping = False
        cls.reader = threading.Thread(target=cls._get_key)
        cls.reader.start()
        cls.started = True

    @classmethod
    def stop(cls):
        if cls.started and cls.reader.is_alive():
            cls.stopping = True
            try:
                cls.reader.join()
            except Exception:
                pass

    @classmethod
    def last(cls) -> str:
        if cls.list:
            return cls.list.pop()
        else:
            return ""

    @classmethod
    def get(cls) -> str:
        if cls.list:
            return cls.list.pop(0)
        else:
            return ""

    @classmethod
    def get_mouse(cls) -> Tuple[int, int]:
        if cls.new.is_set():
            cls.new.clear()
        return cls.mouse_pos

    @classmethod
    def mouse_moved(cls) -> bool:
        if cls.mouse_move.is_set():
            cls.mouse_move.clear()
            return True
        else:
            return False

    @classmethod
    def has_event(cls) -> bool:
        return bool(cls.list)

    @classmethod
    def clear(cls):
        cls.list = []

    @classmethod
    def input_wait(cls, sec: float = 0.0, mouse: bool = False) -> bool:
        """Returns True if key is detected else waits out timer and returns False"""
        if cls.list:
            return True
        # if mouse:
        # Draw.now(Term.mouse_direct_on)
        cls.new.wait(sec if sec > 0 else 0.0)
        # if mouse:
        # Draw.now(Term.mouse_direct_off, Term.mouse_on)

        if cls.new.is_set():
            cls.new.clear()
            return True
        else:
            return False

    @classmethod
    def break_wait(cls):
        cls.list.append("_null")
        cls.new.set()
        sleep(0.01)
        cls.new.clear()

    @classmethod
    def _get_key(cls):
        """Get recive keys.

        Get a key or escape sequence from stdin, convert to readable format
        and save to keys list. Meant to be run in it's own thread.
        """
        input_key: str = ""
        clean_key: str = ""
        try:
            while not cls.stopping:
                with Raw(sys.stdin):
                    # * Wait 100ms for input on stdin then restart loop to check for stop flag
                    if not select([sys.stdin], [], [], 0.1)[0]:
                        continue
                    # * Read 1 key safely with blocking on
                    input_key += sys.stdin.read(1)
                    if (
                        input_key == "\033"
                    ):  # * If first character is a escape sequence keep reading
                        # * Report IO block in progress to prevent Draw functions from getting a IO Block error
                        cls.idle.clear()
                        # Draw.idle.wait()  # * Wait for Draw function to finish if busy
                        # * Set non blocking to prevent read stall
                        with Nonblocking(sys.stdin):
                            input_key += sys.stdin.read(20)
                            if input_key.startswith("\033[<"):
                                _ = sys.stdin.read(1000)
                        cls.idle.set()  # * Report IO blocking done
                    # errlog.debug(f'{repr(input_key)}')
                    if input_key == "\033":
                        clean_key = (
                            "escape"  # * Key is "escape" key if only containing \033
                        )
                    # * Detected mouse event
                    elif input_key.startswith(
                        ("\033[<0;", "\033[<35;", "\033[<64;", "\033[<65;")
                    ):
                        try:
                            cls.mouse_pos = (
                                int(input_key.split(";")[1]),
                                int(input_key.split(";")[2].rstrip("mM")),
                            )
                        except Exception:
                            pass
                        else:
                            # * Detected mouse move in mouse direct mode
                            if input_key.startswith("\033[<35;"):
                                cls.mouse_move.set()
                                cls.new.set()
                            # * Detected mouse scroll up
                            elif input_key.startswith("\033[<64;"):
                                clean_key = "mouse_scroll_up"
                            # * Detected mouse scroll down
                            elif input_key.startswith("\033[<65;"):
                                clean_key = "mouse_scroll_down"
                            # * Detected mouse click release
                            elif input_key.startswith(
                                "\033[<0;"
                            ) and input_key.endswith("m"):
                                if False:
                                    clean_key = "mouse_click"
                                else:
                                    for (
                                        key_name,
                                        positions,
                                    ) in (
                                        cls.mouse.items()
                                    ):  # * Check if mouse position is clickable
                                        if list(cls.mouse_pos) in positions:
                                            clean_key = key_name
                                            break
                                    else:
                                        clean_key = "mouse_click"
                    elif input_key == "\\":
                        clean_key = "\\"  # * Clean up "\" to not return escaped
                    else:
                        for (
                            code
                        ) in (
                            cls.escape.keys()
                        ):  # * Go trough dict of escape codes to get the cleaned key name
                            if input_key.lstrip("\033").startswith(code):
                                clean_key = cls.escape[code]
                                break
                        else:  # * If not found in escape dict and length of key is 1, assume regular character
                            if len(input_key) == 1:
                                clean_key = input_key
                    if clean_key:
                        # * Store up to 10 keys in input queue for later processing
                        cls.list.append(clean_key)
                        if len(cls.list) > 10:
                            del cls.list[0]
                        clean_key = ""
                        cls.new.set()  # * Set threading event to interrupt main thread sleep
                    input_key = ""

        except Exception as e:
            LOG.error(f"Input thread failed with exception: {e}")
            cls.idle.set()
            cls.list.clear()
            # clean_quit(1, thread=True)



import os
import re
import typing
import logging

from .. import __HOME__, __FUNGITDIR__
from ..commands import run_cmd, run_cmd_with_resp
from .gitoptions import GIT_OPTIONS
from .shared import okay, warn, echo, exit_

LOG = logging.getLogger(__name__)


_TEMPLATE_ZSH = """\
#compdef g

complete_g(){
local curcontext="$curcontext" state line ret=1
typeset -A opt_args

_alternative\\
  \'args:options arg:((\\
%s
  ))\'\\
  'files:filename:_files'
return ret
}

compdef complete_g g
"""

_TEMPLATE_BASH = """\
#!/usr/env bash

_complete_g(){
  if [[ "${COMP_CWORD}" == "1" ]];then
    COMP_WORD="%s"
    COMPREPLY=($(compgen -W "$COMP_WORD" -- ${COMP_WORDS[${COMP_CWORD}]}))
  fi
}

complete -F _complete_g g
"""

_re = re.compile(r"\/\.config\/fungit/([^\s]+)")


def get_current_shell() -> str:
    """Gets the currently used shell.

    Returns:
        shell_: Current shell string.
    """
    shell_ = ""
    _, resp = run_cmd_with_resp("echo $SHELL")
    if resp:
        shell_ = resp.split("/")[-1].strip()
    return shell_


def ensure_config_path(file_name: str) -> str:
    """Check config path.

    Check whether the configuration directory exists, if not, try to create
    it. Failed to exit, successfully returned to complete the file path.

    Args:
        file_name: Completion prompt script name.

    Returns:
        file_path: Full path of completion prompt script.
    """
    LOG.debug(f"{__FUNGITDIR__}, {file_name}")
    if not os.path.exists(__FUNGITDIR__):
        try:
            os.mkdir(__FUNGITDIR__)
            return "{}/{}".format(__FUNGITDIR__, file_name)
        except Exception as e:
            exit_(1, e)
    else:
        file_path = "{}/{}".format(__FUNGITDIR__, file_name)
        return file_path


def generate_complete_script(template: str, fn: typing.Callable, name: str = "_g"):
    """Generate completion scirpt.

    Generate the completion script of the corresponding shell according to
    the template.

    Args:
        template: Script template.
        fn: Method of generating script content.
        name: Completion script name.
    """
    complete_src = fn()
    script_src = template % (complete_src)

    try:
        with open("./%s" % (name), "w") as f:
            for line in script_src:
                f.write(line)
    except Exception as e:
        exit_(1, e)


def using_completion(file_name: str, path: str, config_path: str):
    """Try using completion script.

    Inject the load of completion script into the configuration of shell.
    If it exists in the configuration, the injection will not be repeated.

    Args:
        file_name: generated completion script.
        path: `fungit` configuration path.
        config_path: shell configuration path.
    """
    try:
        run_cmd("mv {} {}".format(file_name, __FUNGITDIR__))
        with open(config_path) as f:
            shell_conf = f.read()
            files = _re.findall(shell_conf)
    except Exception as e:
        exit_(1, e)

    has_injected = False
    if files:
        for file in files:
            if file == file_name:
                has_injected = True
    LOG.debug(f"has_injected: {has_injected}")

    if not has_injected:
        try:
            run_cmd('echo "source %s" >> %s ' % (path, config_path))
        except Exception as e:
            exit_(1, e)
        okay("\nPlease run: source {}".format(config_path))
    else:
        warn("This configuration already exists.")


def add_zsh_completion():
    """Add Zsh completion prompt script."""

    _name = "_g"
    _path = ensure_config_path(_name)

    def gen_completion():
        vars = []

        for k in GIT_OPTIONS.keys():
            desc = GIT_OPTIONS[k]["help-msg"]
            if not desc:
                desc = "no description."
            vars.append('    {}\\:"{}"\\\n'.format(k, desc))

        return ("\n".join(vars)).rstrip()

    generate_complete_script(_TEMPLATE_ZSH, gen_completion, _name)

    using_completion(_name, _path, __HOME__ + "/.zshrc")


def add_bash_completion():
    """Add Bash completion prompt script."""

    _name = "complete_script"
    _path = ensure_config_path(_name)

    def gen_completion():
        return " ".join(GIT_OPTIONS.keys())

    generate_complete_script(_TEMPLATE_BASH, gen_completion, _name)

    using_completion(_name, _path, __HOME__ + "/.bashrc")


def add_completion():
    """Add completion prompt script."""
    echo("\nTry to add completion ...")

    _shell = get_current_shell()
    echo("Detect shell: %s" % _shell)
    if _shell == "zsh":
        add_zsh_completion()
    elif _shell == "bash":
        add_bash_completion()
    else:
        warn("Don't support completion of %s" % _shell)



#!/usr/bin/python3
import argparse
import platform
import signal
import logging

from . import __version__, __license__, __author__, __git_url__
from .logutil import setup_logging
from .term import Term
from .gui.box_option import generate_all_box
from .event.key import Key
from .event.process import process_key
from .event.clean_quit import quit_app


LOG = logging.getLogger(__name__)


def main(costom_commands: list = []):
    args = argparse.ArgumentParser()
    args.add_argument(
        "-v", "--version", action="store_true", help="show version info and exit"
    )
    args.add_argument(
        "--debug",
        action="store_true",
        help="start with loglevel set to DEBUG overriding value set in config",
    )
    stdargs = args.parse_args()
    if costom_commands:
        stdargs = args.parse_args(costom_commands)

    if stdargs.version:
        print(
            f"[fungit]: {__version__}, license={__license__}, os={platform.system()}, arch={platform.machine()}"
        )
        print(f"author={__author__}, url={__git_url__}")
        raise SystemExit(0)

    DEBUG: bool = stdargs.debug

    # Mark signal
    try:
        signal.signal(signal.SIGINT, signal_quit)  # Ctrl + c
        signal.signal(signal.SIGWINCH, signal_resize)  # Terminal resized
    except Exception as e:
        print("Error: initialize single failed! -> %s" % e)
        quit_app()

    setup_logging(DEBUG)
    Term.init()
    Key.start()
    generate_all_box(recreate=True)

    def run():
        while not False:
            # Term.refresh()
            # Timer.stamp()

            # while Timer.not_zero():
            #     if Key.input_wait(Timer.left()):
            #         process_key()
            process_key()

    try:
        run()
    except Exception as e:
        LOG.error(e)
        quit_app()


def signal_resize(signum, frame):
    generate_all_box(recreate=True, update_data=False)


def signal_quit(signum, frame):
    quit_app()



class BoxType:
    STATE = 1
    STATUS = 1 << 1
    COMMIT = 1 << 2
    BRANCH = 1 << 3
    STASH = 1 << 4
    CONTENT = 1 << 5
    Tip = 1 << 6
    Input = 1 << 7


class GitActionStatus:
    NONE = 1
    PULLING = 1 << 1


class ConfirmType:
    NORMAL = 1
    ERROR = 1 << 1



from typing import Any

from ..shared import BoxType
from ..utils import create_profile
from ..renderer import Renderer


class Box:
    current: BoxType
    name: str
    genre: int
    t_w: int
    t_h: int
    x: int
    y: int
    w: int
    h: int
    content: Any
    box: str

    @classmethod
    def create_profile(cls):
        # _line_color = BOX_SELECTED_COLOR if cls.genre & cls.current else ''
        _line_color = ""
        cls.box = create_profile(
            cls.x, cls.y, cls.w, cls.h, cls.name, line_color=_line_color
        )

    @classmethod
    def render(cls):
        Renderer.now(cls.box, cls.box_content)



import sys

PYTHON_VERSION = sys.version_info[:2]
if PYTHON_VERSION < (3, 7):
    print("Python version must be greater than or equal to 3.7")
    sys.exit()

try:
    LONG_DESCRIPTION = open("README.md").read()
except Exception:
    LONG_DESCRIPTION = ""

from setuptools import setup, find_packages
import fungit

setup(
    name="fungit",
    version=fungit.__version__,
    author=fungit.__author__,
    author_email=fungit.__email__,
    description="Simple terminal tool of Git.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url=fungit.__git_url__,
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[],
    entry_points="""
        [console_scripts]
        fungit=fungit.main:main
        g=fungit.terminal_git.main:command_g
    """,
    python_requires=">=3.7",
)



from typing import Tuple

from fungit.style import Symbol, Fx, Color, Cursor


def create_profile(
    x: int = 0,
    y: int = 0,
    width: int = 0,
    height: int = 0,
    title: str = "",
    title2: str = "",
    line_color: Color = None,
    title_color: Color = None,
    fill: bool = True,
    box=None,
) -> str:
    """Create a box from a box object or by given arguments"""
    # out: str = f'{Term.fg}{Term.bg}'
    out: str = ""
    if not line_color:
        line_color = ""
    if not title_color:
        title_color = ""

    # * Get values from box class if given
    if box:
        x = box.x
        y = box.y
        width = box.width
        height = box.height
        title = box.name
    h_lines: Tuple[int, int] = (y, y + height - 1)

    out += f"{line_color}"

    # * Renderer all horizontal lines
    for h_pos in h_lines:
        out += f"{Cursor.to(h_pos, x)}{Symbol.h_line * (width - 1)} "

    # * Renderer all vertical lines and fill if enabled
    for h_pos in range(h_lines[0] + 1, h_lines[1]):
        # out += f'{Cursor.to(h_pos, x)}{Symbol.v_line}{" " * (width-2) if fill else Cursor.r(width-2)}{Symbol.v_line}'
        out += "%s%s%s%s" % (
            Cursor.to(h_pos, x),
            Symbol.v_line,
            " " * (width - 2) if fill else Cursor.r(width - 2),
            Symbol.v_line,
        )

    # * Renderer corners
    if height > 1:
        out += "%s%s%s%s%s%s%s%s" % (
            Cursor.to(y, x),
            Symbol.left_up,
            Cursor.to(y, x + width - 1),
            Symbol.right_up,
            Cursor.to(y + height - 1, x),
            Symbol.left_down,
            Cursor.to(y + height - 1, x + width - 1),
            Symbol.right_down,
        )

    # * Renderer titles if enabled
    if title:
        numbered: str = ""
        out += "%s%s%s%s%s%s%s%s%s" % (
            Cursor.to(y, x + 2),
            Symbol.title_l,
            Fx.b,
            numbered,
            title_color,
            title,
            Fx.ub,
            line_color,
            Symbol.title_r,
        )
    if title2:
        # out += f'{Cursor.to(hlines[1], width - 2 - len(title2))}{Symbol.title_left}{title_color}{Fx.b}{title2}{Fx.ub}{line_color}{Symbol.title_right}'
        out += f"{Cursor.to(h_lines[1], width - 2 - len(title2))}{title_color}{Fx.b}{title2}{Fx.ub}{line_color}"

    # return f'{out}{Term.fg}{Cursor.to(y + 1, x + 1)}'
    return f"{out}{Fx.reset}{Cursor.to(y + 1, x + 1)}"



import logging

from fungit.commands import run_cmd_with_resp
from .gitoptions import GIT_OPTIONS, TYPES
from .shared import echo, warn, err, exit_, CommandColor, Fx


LOG = logging.getLogger(__name__)


def echo_one_help_msg(k: str):
    """Print a tip.

    Find the corresponding help information according to the `k` value and
    print it. If the help information does not exist, print the executed
    full command.

    Args:
        k: Short command.
    """
    echo("    " + k, color=CommandColor.GREEN, nl=False)
    if GIT_OPTIONS[k]["help-msg"]:
        msg = GIT_OPTIONS[k]["help-msg"]
    else:
        msg = GIT_OPTIONS[k]["command"]
    echo((9 - len(k)) * " " + str(msg))


def echo_help_msg():
    """Print help message."""
    echo("These are short commands that can replace git operations:")
    for k in GIT_OPTIONS.keys():
        echo_one_help_msg(k)


def give_tip(t: str):
    """Print a part of help message.

    Print the help information of the corresponding part according to the
    incoming command type string. If there is no print error prompt for the
    type.

    Args:
        t: A command type.
    """
    t = t[0].upper() + t[1:].lower() if len(t) > 2 else ""
    if t not in TYPES:
        err("There is no such type")
        raise SystemExit(0)

    echo(f"These are the orders of {t}")
    prefix = t[0].lower()
    for k in GIT_OPTIONS.keys():
        if k.startswith(prefix):
            echo_one_help_msg(k)


def echo_types():
    """Print all command types."""
    for t in TYPES:
        print(f" {t}")


def echo_description():
    """Print the description information"""
    from .. import __version__

    has_git = False
    try:
        _, git_version = run_cmd_with_resp("git --version")
        if git_version:
            has_git = True
    except Exception:
        LOG.warning("Happen error when run command with get Git version")
        git_version = ""

    echo("[fungit] version: %s" % __version__, style=Fx.b)
    echo(git_version)
    echo("Description:", style=Fx.b)
    echo(
        (
            "Fungit terminal tool, help you use git more simple."
            " Support Linux and MacOS.\n"
        ),
        style=Fx.underline,
    )

    echo("You can use ", nl=False)
    echo("-h", color=CommandColor.GREEN, nl=False)
    echo(" and ", nl=False)
    echo("--help", color=CommandColor.GREEN, nl=False)
    echo(" to get how to use command fungit.\n")

    if not has_git:
        warn("Don't found Git, maybe need install.")
